/* $Id: common.c,v 1.1.1.1 2004/11/06 01:16:47 x86man Exp $ */

#include "config.h"
#include "platform.h"
#include "structs.h"


/*

#ifdef OS_AIX
void *rpl_malloc(size_t n) { return malloc(n); }
#endif

*/

#ifdef BUILDING_DAEMON

#include "neyed.h"

#endif

#ifdef HAVE_LIBPTHREAD

pthread_mutex_t protocol_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t memmgr_mutex = PTHREAD_MUTEX_INITIALIZER;

#endif

static unsigned long malloc_total = 0L;
static unsigned malloc_counts = 0;
static unsigned free_counts = 0;

void IpProtocol(unsigned char proto, char *strprotocol)
{
	char strmsg[STDSTRSIZE];
        struct protoent *p;

        assert(strprotocol != NULL);

#ifdef HAVE_LIBPTHREAD
/*	 while (pthread_mutex_trylock(&protocol_mutex) != 0)
	 {
		WaitState();
	 }
*/
#endif

        p = (struct protoent *) getprotobynumber((int)proto);


        if (p == NULL)
        {
                sprintf(strprotocol,"Unknown");
		sprintf(strmsg,"Your /etc/protocols has gone bitches, please fix it since protocol %d is unknown !",proto);
		notify(strmsg);
        }
        else
        {
                sprintf(strprotocol,"%s",p->p_name);
        }

#ifdef HAVE_LIBPTHREAD
/*
	pthread_mutex_unlock(&protocol_mutex);
*/
#endif
}

void RemoveCR(char *string)
{
	int len,i;
	if (string == NULL)
		return;

	len=strlen(string);

	for (i=(len-1); i>=0; i--)
	{
		if (string[i] == '\n' || string[i] == '\r')
		{
			string[i] = 0;
			break;
		}
	}
	
}

#ifndef MEMWATCH

void InitMemoryManager()
{
	malloc_counts = 0;
	malloc_total  = 0L;
	free_counts = 0;
#ifdef HAVE_LIBPTHREAD
	pthread_mutex_init(&memmgr_mutex,NULL);
#endif
}

void *allocate_mem(size_t size)
{
	void *ptr;
#ifdef HAVE_LIBPTHREAD
	while (pthread_mutex_trylock(&memmgr_mutex) != 0)
	{
		WaitState();
	}
#endif

	ptr = malloc(size);
	if (ptr == NULL)
	{
		fprintf(stderr,"Can't allocate %d bytes\n",size);
		exit(1);
	}
#ifdef DEBUG
	memset(ptr,0,size);
#endif
	malloc_counts++;
	malloc_total += (unsigned long)size;
#ifdef HAVE_LIBPTHREAD
	pthread_mutex_unlock(&memmgr_mutex);
#endif
	return ptr;
}

void release_mem(void *ptr,size_t size)
{
	assert(ptr!=NULL);
#ifdef HAVE_LIBPTHREAD
	while (pthread_mutex_trylock(&memmgr_mutex) != 0)
	{
		WaitState();
	}
#endif
#ifdef DEBUG
	memset(ptr,0,size);
#endif
	free(ptr);
	free_counts++;
	malloc_total-=(unsigned long)size;
#ifdef HAVE_LIBPTHREAD
	pthread_mutex_unlock(&memmgr_mutex);
#endif
}

#endif

void notify(char *msg)
{
	fprintf(stderr,"%s\n",msg);
	fflush(stderr);
	syslog(LOG_DAEMON,msg);
}

void trace(char *msg)
{
        fprintf(stderr,"%s\n",msg);
	fflush(stderr);
}


/* #ifdef BUILDING_DAEMON */


static unsigned long processed   = 0;
static unsigned long encountered = 0;
static unsigned long ReceivedPDU = 0;

void InitQueueCounters()
{
	processed = encountered = ReceivedPDU = 0;
}

unsigned long GetProcessedCount()
{
	return processed;
}

unsigned long GetQueueCount()
{
	return encountered;
}

unsigned long GetReceivedCount()
{
	return ReceivedPDU;
}

void IncrementProcessed()
{
	processed++;
}

void IncrementQueued()
{
	encountered++;
}

void IncrementReceived()
{
	ReceivedPDU++;
}

void ResetQueued()
{
	encountered = 0;
}

void ShowDebugger()
{
	char msg[STDSTRSIZE];
	sprintf(msg,"\nDEBUG: General - Mallocs: %ld  - Free: %ld - To be freed: %ld",malloc_counts,free_counts,malloc_counts-free_counts);
	trace(msg);
	sprintf(msg,"DEBUG: Total allocated memory: %ld\n",malloc_total);
	trace(msg);
	sprintf(msg,"DEBUG: Dispatcher - Received: %ld - Queued: %ld - Processed: %ld",
				GetReceivedCount(),
				GetQueueCount(),
				GetProcessedCount());
	trace(msg);
	sprintf(msg,"DEBUG: Storage - Written: %ld - Freed: %ld ",
				GetStorageWrittenCount(),
				GetStorageFreedCount());
	trace(msg);
}

/* #endif  */
