/* $Id: entry.c,v 1.2 2004/11/06 03:16:56 x86man Exp $ */

#include "config.h"
#include "neyed.h"
#include "getopt.h"

#if defined(OS_ULTRIX) || defined(OS_NEXTSTEP)
#include "strdup.c"
#endif

#if defined(OS_ULTRIX) || defined(OS_IRIX) || defined(OS_AIX) || defined(OS_HPUX)
#include "getopt.c"
#endif

static char Release[] = VERSION;

static char *TextLogFile;
static char *BinaryLogFile;
static char *TrashLogFile;
static char *cBindAddress;
static int nTextLogEnabled;
static int nBinaryLogEnabled;
static int nTrashLogEnabled;
static int nDetached;
#ifdef HAVE_LIBPTHREAD
static unsigned long nMaxThrottle;
#endif 

#ifdef HAVE_LIBMYSQLCLIENT
static int nMYSQLLogEnabled;
static char *MYSQLServerName;
#endif

#ifdef HAVE_GETPWNAM
#include "pwd.h"
#endif

char *GetBinaryLogFileName()
{
	return BinaryLogFile;
}
char *GetTextLogFileName()
{
	return TextLogFile;
}

char *GetTrashLogFileName()
{
	return TrashLogFile;
}

int IsDetached()
{
	return nDetached;
}

int IsTrashLoggingEnabled()
{
	return nTrashLogEnabled;
}

int IsTextLoggingEnabled()
{
	return nTextLogEnabled;
}

int IsBinaryLoggingEnabled()
{
	return nBinaryLogEnabled;
}

#ifdef HAVE_LIBPTHREAD
unsigned long GetMaxThrottle()
{
	return nMaxThrottle;
}
#endif 

#ifdef HAVE_LIBMYSQLCLIENT

int IsMYSQLLoggingEnabled()
{
	return nMYSQLLogEnabled;
}


char *GetMYSQLServerName()
{
	return MYSQLServerName;
}

#endif
void ReleaseNamedVars()
{
	if (cBindAddress != NULL)
		release_mem(cBindAddress,strlen(cBindAddress));
	if (TextLogFile != NULL)
		release_mem(TextLogFile,strlen(TextLogFile));
	if (BinaryLogFile != NULL)
		release_mem(BinaryLogFile,strlen(BinaryLogFile));
	if (TrashLogFile != NULL)
		release_mem(TrashLogFile,strlen(TrashLogFile));
#ifdef HAVE_LIBMYSQLCLIENT
	if (MYSQLServerName != NULL)
		release_mem(MYSQLServerName,strlen(MYSQLServerName));
#endif
}

void SwitchTrashFlag()
{
	if (nTrashLogEnabled == 1)
	{
		nTrashLogEnabled =0;
	}
	else
	{
		nTrashLogEnabled = 1;
	}
	SetupSignals();
}

void SwitchTextFlag()
{
	if (nTextLogEnabled == 1)
	{
		nTextLogEnabled =0;
	}
	else
	{
		nTextLogEnabled = 1;
	}
	SetupSignals();
}

void HandleSnapshotSignal()
{
#ifdef HAVE_LIBSQLITE
	CreateSnapshot();
#endif
	SetupSignals();	
}

void DoRotateDB()
{
#ifdef HAVE_LIBSQLITE
	RotateDB();
#endif
	SetupSignals();
}

void SetupSignals()
{
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	printf("I'm in the SetupSignals()\n");
#endif
#endif
	signal(SIGHUP,&SwitchTrashFlag);
#ifdef HAVE_LIBSQLITE	
	signal(SIGUSR1,HandleSnapshotSignal);
#endif
	signal(SIGUSR2,SwitchTextFlag);
	signal(SIGWINCH,DoRotateDB);
	signal(SIGTERM,HaltApp);
} 

char *OSVersion()
{
	return strdup(OSNAME);
}

void HelpCommandLine()
{
	fprintf(stdout,"NetFlow Traffic Collector %s\n",Release);
	fprintf(stdout,"License: BITGPL, by Gilberto Persico\n");
#ifdef TELCO_SUX
	fprintf(stdout,"Yep, running this software I agree that Italian telcos are suxing thieves !\n");
#else
	#error "You are faking license, thief !"
#endif
#ifdef HAVE_LIBPTHREAD
	fprintf(stdout,"Multithreaded version\n"); 
	fprintf(stdout,"Running on: %s\n",OSVersion());
#else
	fprintf(stdout,"Single Thread version\n"); 
	fprintf(stdout,"Running on: %s\n",OSVersion());
#endif
	fprintf(stdout,"Command Line switches are :\n");
	fprintf(stdout,"-d Daemonize and go background\n");
	fprintf(stdout,"-p UDP Port on which NetFlows are incoming\n");
	fprintf(stdout,"-a IP Address to which NetFlows are incoming\n");
	fprintf(stdout,"-l Enable ASCII logging, defaults to %s\n",DEFAULT_TEXTLOGFILE);
#ifdef HAVE_LIBSQLITE
	fprintf(stdout,"-b Enable binary logging, defaults to %s\n",DEFAULT_BINARYLOGFILE);
#endif
#ifdef HAVE_LIBMYSQLCLIENT
	fprintf(stdout,"-m ip/hostname of MYSQL database server\n");
#endif
	fprintf(stdout,"-t Enable trash (non NetFlow5 packets) logging, defaults to %s\n",DEFAULT_TRASHLOGFILE);
	fprintf(stdout,"-x Disable text logging\n");
#ifdef HAVE_LIBSQLITE
	fprintf(stdout,"-y Disable binary logging\n");
#endif
	fprintf(stdout,"-z Disable trash logging\n");
#ifdef HAVE_GETPWNAM
	fprintf(stdout,"-u Run neyed as user xxxx\n");
#endif

#ifdef HAVE_LIBPTHREAD
	fprintf(stdout,"-n Threads to use on multithreaded versions (default: %d)\n",DEFAULT_NTHREADS);
	fprintf(stdout,"-i Throttle limit - When buffer queue reaches this limit, give packet processing maximum priority (default: %d packets)\n",DEFAULT_THROTTLE);
#endif
	fprintf(stdout,"SIGNALS Available:\n");
#ifdef HAVE_LIBSQLITE
	fprintf(stdout,"SIGUSR1 - Creates a snapshot of the current DB\n");
#endif
	fprintf(stdout,"SIGUSR2 - Enables/Disables ASCII logging\n");
	fprintf(stdout,"SIGTERM - Performs a clean shutdown\n");
	fprintf(stdout,"SIGHUP - Enables/Disables Trash logging\n");
#ifdef HAVE_LIBSQLITE
	fprintf(stdout,"SIGWINCH - Rotate current DB\n");
#endif
	fflush(stdout);
	exit(1);
}


int main(int argc, char **argv)
{
	static int RetVal;
	static int nThreads;
	static int nNetflowport;
	static int arg;
	static long len;
#ifdef HAVE_LIBPTHREAD
	static char ErrMsg[STDSTRSIZE];
#endif
#ifdef HAVE_GETPWNAM
	static char *user;
	struct passwd *pw;
	uid_t myuid;
#endif

#ifdef DMALLOC
	dmalloc_debug_setup("debug=0x44d4cd03,inter=100,log=/tmp/dmalloc.log");
#endif

	InitMemoryManager();

	cBindAddress = TextLogFile = BinaryLogFile = TrashLogFile = NULL;

	nTextLogEnabled = 1;
	nBinaryLogEnabled = 1;
	nTrashLogEnabled = 1;
	nThreads = DEFAULT_NTHREADS;

	nDetached = 0;
	user = NULL;

	nNetflowport = NETFLOWPORT;
#ifdef HAVE_LIBMYSQLCLIENT
	nMYSQLLogEnabled = 0;
#endif

#ifdef HAVE_LIBPTHREAD
	nMaxThrottle = DEFAULT_THROTTLE;
	EnableWaitState();
#endif

	while ((arg=getopt(argc,argv,"p:a:l:b:t:n:m:hxyzdi:u:")) != EOF)
	{
		switch (arg)
		{
			case 'p':
			{
				nNetflowport=atoi(argv[optind-1]);
				break;
			}
#ifdef HAVE_GETPWNAM
			case 'u':
			{
				len = strlen(argv[optind-1]);
				if (len > 0 && len < 9 )
				{
					user = strdup(argv[optind-1]);
				}
				break;
			}
#endif
#ifdef HAVE_LIBPTHREAD
			case 'n':
			{
				nThreads=atoi(argv[optind-1]);
				break;
			}
			case 'i':
			{
				nMaxThrottle=atol(argv[optind-1]);
				break;
			}
#endif
			case 'a':
			{
				cBindAddress = (char *)strdup(argv[optind-1]);
				break;
			}
#ifdef HAVE_LIBMYSQLCLIENT		
			case 'm':
			{
				MYSQLServerName = (char *)strdup(argv[optind-1]);
				nMYSQLLogEnabled = 1;
				break;
			}		
#endif			
			case 'l':
			{
				TextLogFile = (char *)strdup(argv[optind-1]);
				break;
			}
#ifdef HAVE_LIBSQLITE
			case 'b':
			{
				BinaryLogFile = (char *)strdup(argv[optind-1]);
				break;
			}
#endif
			case 't':
			{
				TrashLogFile = (char *)strdup(argv[optind-1]);
				break;
			}
			case 'x':
			{
				nTextLogEnabled = 0;
				break;
			}
#ifdef HAVE_LIBSQLITE
			case 'y':
			{
				nBinaryLogEnabled = 0;
				break;
			}
#endif
			case 'd':
			{
				nDetached = 1;
				break;
			}
			case 'z':
			{
				nTrashLogEnabled = 0;
				break;
			}
			case 'h':
			{
				HelpCommandLine();
				break;
			}
		}
	}

	if (cBindAddress == NULL)
	{
		cBindAddress = (char *)strdup(LISTENADDR);
	}

	if (TextLogFile == NULL)
	{
		TextLogFile  = (char *)strdup(DEFAULT_TEXTLOGFILE);
	}

	if (TrashLogFile == NULL)
	{
		TrashLogFile  = (char *)strdup(DEFAULT_TRASHLOGFILE);
	}

	if (BinaryLogFile == NULL)
	{
		BinaryLogFile  = (char *)strdup(DEFAULT_BINARYLOGFILE);
	}

#ifdef HAVE_GETPWNAM
	if (user != NULL)
	{
		pw = getpwnam(user);
		if (pw == NULL)
		{
			fprintf(stdout,"User %s does not exist !\n");
			exit(1);
		}
		myuid = pw->pw_uid;
		setuid(myuid);	
		release_mem(user,strlen(user));
	}
#endif

	SetupSignals();


	fprintf(stdout,"Netflow listener starting with PID: %ld\n",(long)getpid());
	fprintf(stdout,"Listening on %s:%d\n",cBindAddress,nNetflowport);
	fprintf(stdout,"ASCII logging: %s\n", (nTextLogEnabled==0)?"Disabled":TextLogFile);
#ifdef HAVE_LIBSQLITE
	fprintf(stdout,"Binary logging (SQLITE %s): %s\n", sqlite_version, (nBinaryLogEnabled==0)?"Disabled":BinaryLogFile);
#endif
	fprintf(stdout,"Trash logging: %s\n", (nTrashLogEnabled==0)?"Disabled":TrashLogFile);
#ifdef HAVE_LIBMYSQLCLIENT
	if (IsMYSQLLoggingEnabled()== 1)
	{
		fprintf(stdout,"DB: Will store on MYSQL database %s on %s\n",NETFLOWDB,GetMYSQLServerName());
	}
#endif
#ifdef HAVE_LIBPTHREAD
	fprintf(stdout,"Max Throttle at %ld packets\n",GetMaxThrottle());
#endif
	/*********************************************/
	/* Here we check we can access the files ... */
	/*********************************************/

	if (nTextLogEnabled != 0)
	{
		if (access(TextLogFile,W_OK) != 0)
		{
			if (access(TextLogFile,F_OK) == 0)
			{
				Terminate(ERR_FILEACCESS,
					"Main, can't access to Text Log File, check permissions");
			}
			LogWrite("Begin Logging\n",(char *)GetTextLogFileName());
		}
	}
#ifdef HAVE_LIBSQLITE
	if (nBinaryLogEnabled != 0)
	{
		if (access(BinaryLogFile,W_OK) != 0)
		{
			if (access(BinaryLogFile,F_OK) == 0)
			{
				Terminate(ERR_FILEACCESS,
					"Main, can't access to Binary Log File, check permissions");
			}
		}
	}
#endif
	if (nTrashLogEnabled != 0)
	{
		if (access(TrashLogFile,W_OK) != 0)
		{
			if (access(TrashLogFile,F_OK) == 0)
			{
				Terminate(ERR_FILEACCESS,
					"Main, can't access to Trash Log File");
			}
			LogWrite("Begin Logging\n",(char *)GetTrashLogFileName());
		
		}
	}

	/* Here we check our own limits ... */

#ifdef HAVE_LIBPTHREAD
	if (nThreads <= 0  || nThreads > MAX_THREADS)
	{
		sprintf(ErrMsg,"Main - The number of threads must be a number between 1 and %d !",MAX_THREADS);
		Terminate(ERR_THREADCOUNT,ErrMsg);
	}
	fprintf(stdout,"Will run with %d threads\n",nThreads);
#endif
#ifdef HAVE_GETPWNAM
	if (user != NULL)
	{
		fprintf(stdout,"Running as user %s\n",user);
	}
	else
	{
		pw = getpwuid(getuid());
		if (pw != NULL)
		{
			fprintf(stdout,"Running as user %s\n",pw->pw_name);
		}
		else	
		{
			fprintf(stdout,"Running as unknown user !\n");
		}
	}
#endif

	RetVal = ListenerProcess(nNetflowport,cBindAddress,nThreads);


	return RetVal;
}
