/* $Id: listen.c,v 1.1.1.1 2004/11/06 01:16:47 x86man Exp $ */

#include "config.h"
#include "neyed.h"

#ifdef HAVE_LIBPTHREAD
#include "jobs.h"
#endif

static int ShuttingDown = 0;

void StartShutdown()
{
	ShuttingDown = 1;
}

int ShutdownInProgress()
{
	return ShuttingDown;
}

static struct sockaddr_in NFaddr;

static int NFSocketDescriptor;

int ListenerProcess(int netflowport, char * listenaddress, int nThreads)
{
	static int RetVal;
	int on = 1;

#ifdef FUNCTRACE
	printf("I'm in the ListenerProcess(%d,%s)\n",
					netflowport,
					listenaddress); 
#endif

	NFSocketDescriptor = socket(PF_INET, SOCK_DGRAM,0);

	bzero(&NFaddr, sizeof(NFaddr));

	NFaddr.sin_family = AF_INET;
	NFaddr.sin_port   = htons(netflowport);
	NFaddr.sin_addr.s_addr = inet_addr(listenaddress);


#ifdef SO_REUSEADDR
	setsockopt(NFSocketDescriptor,SOL_SOCKET,SO_REUSEADDR,(char *)&on,sizeof(on));
#endif 
#ifdef OS_SOLARIS
	on=262144;
#else
	on=1048576 * 4;
#endif
	setsockopt(NFSocketDescriptor,SOL_SOCKET,SO_RCVBUF,(char *)&on,sizeof(on));

/*	fcntl(NFSocketDescriptor, F_SETFL, O_NONBLOCK); */


	RetVal = bind(NFSocketDescriptor, (struct sockaddr *)&NFaddr, sizeof(NFaddr));
	if (RetVal != 0)
	{
		Terminate(ERR_CANTBIND,"ListenProcess - can't bind to port");
	}

	return NeverEndingLoop(nThreads);
}

int NeverEndingLoop(int nThreads)
{
	int maxfd = 0;
	fd_set filedescriptors;
	int DataInQueue = 0;
	struct timeval timeout;
	char *packet;
	int packet_len;
	struct sockaddr from;
	int from_len;
#ifdef DEBUG					
	long incoming;
#endif	

#ifdef HAVE_LIBPTHREAD
	int RetVal;
	pthread_t	dispatcher;
	pthread_attr_t	dispatcher_attribute;
	pthread_t tid;
	pthread_attr_t thread_attribute;
	int count;

#endif
		
#ifdef FUNCTRACE
	printf("I'm in the NeverEndingLoop()\n"); 
	printf("FDs are NF: %d\n",NFSocketDescriptor);
#endif

	if (IsDetached() == 1)
	{
		switch (fork())
		{
			case -1:
			{
				Terminate(ERR_SHUTDOWN,"Clean shutdown\n");
			}
			case 0:
			{
				setsid();
				fprintf(stderr,"Succesfully daemonized with pid %ld\n",(long)getpid());
				fflush(stderr);
				freopen("/dev/null", "r", stdin);
                                freopen("/dev/null", "w", stdout);
                                freopen("/dev/null", "w", stderr);				
				break;
			}
			default:
			{
				exit(0);
			}
		}
	}


#ifdef HAVE_LIBPTHREAD

	InitThreadList();


#ifdef OS_SOLARIS
        pthread_setconcurrency(nThreads);
#endif


	pthread_attr_init(&dispatcher_attribute);
/*
	pthread_attr_setscope(&dispatcher_attribute,PTHREAD_SCOPE_SYSTEM);
*/
	RetVal = pthread_create(&dispatcher,
				&dispatcher_attribute, 
				(void *)&StorageJob,
				NULL);
	if (RetVal != 0)
	{
		Terminate(ERR_THREADCREAT,"NeverendingLoop - Can't create storage thread");
	}


	AddThread(dispatcher, THREAD_ALIVE, THREAD_TYPE_STORAGE,-1);

	pthread_attr_init(&dispatcher_attribute);

	RetVal = pthread_create(&dispatcher,
				&dispatcher_attribute, 
				(void *)&JobListMaintainer,
				(void *)nThreads);
	if (RetVal != 0)
	{
		Terminate(ERR_THREADCREAT,"NeverendingLoop - Can't create dispatcher thread");
	}
	
	AddThread(dispatcher, THREAD_ALIVE, THREAD_TYPE_DISPATCHER,-1);

	SetThreadCount(nThreads);
	SetWaitTime(nThreads);
	
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{		
		fprintf(stderr,"Creating %d workers\n",nThreads);
		fflush(stderr);	
	}
#endif


	pthread_attr_init(&thread_attribute);
/*
	pthread_attr_setscope(&thread_attribute, PTHREAD_SCOPE_SYSTEM);
*/

	for (count=0;count<nThreads;count++)
	{
		pthread_create(	&tid,
				&thread_attribute, 
				(void *)&ThreadWorker,
				(void *)count);
		AddThread(tid, THREAD_ALIVE, THREAD_TYPE_WORKER,count);
	}

	
#else
	InitStorage();
#endif

#ifdef DEBUG					
	incoming = 0;
#endif	

#ifdef HAVE_LIBPTHREAD
	SetThrottle(0);
#endif
	packet = allocate_mem(STDBUFSIZE);

	while (ShutdownInProgress() == 0)
	{	

#ifdef UNCERTAIN_DEBUG
		fprintf(stderr,"Waiting for packets ...\n");
		fflush(stderr);
#endif		

		SetupSignals();

		timeout.tv_sec  = 0;
		timeout.tv_usec = 75000;

		FD_ZERO(&filedescriptors);
		maxfd = NFSocketDescriptor;
		FD_SET(NFSocketDescriptor,&filedescriptors);

		DataInQueue = select(maxfd+1,&filedescriptors, 0, 0, &timeout);

		if (DataInQueue > 0 )
		{
#ifdef HAVE_LIBPTHREAD
			SetThrottle(1);
#endif
			if (FD_ISSET(NFSocketDescriptor,&filedescriptors) )
			{
				from_len = sizeof(from);
				packet_len = recvfrom(NFSocketDescriptor,
							packet, 
							STDBUFSIZE, 
							0, 
							&from, 
							&from_len);
				if (packet_len <= 0)
				{
					release_mem(packet,STDBUFSIZE);
				}
				else
				{
#ifdef HAVE_LIBPTHREAD
					AddJob(packet,packet_len,&from);
#ifdef DEBUG					
					incoming++;
					if (IsDetached() == 0)
					{
						fprintf(stderr,"Queued Packets: %09ld Incoming packets: %ld\r",GetQueueCount(),GetReceivedCount()); 
					}				
#endif					
					
#else	
					NetFlowPacketParser(packet,
								&from,
								packet_len);
#endif				
				}
			}
		}
		else
		{
#ifdef HAVE_LIBPTHREAD
			SetThrottle(0);
#endif
#ifdef DEBUG
			ShowDebugger();
#endif
		}
	}
	release_mem(packet,STDBUFSIZE);
	ReleaseNamedVars();
	return 0;
}
