/* $Id: neyed.h,v 1.1.1.1 2004/11/06 01:16:48 x86man Exp $ */

#ifndef __TRAFFICLISTEN__
#define __TRAFFICLISTEN__

#include "defaults.h"

#include "platform.h"


#ifndef max  

#define max(a,b) (a>b)?a:b 
#define min(a,b) (a<b)?a:b

#endif

#include "structs.h"

#ifdef ENABLE_PTHREADS
#include "jobs.h"
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

/***************************************************
Local function headers
****************************************************/
int ListenerProcess(int netflowport, char * listenaddress,int nThreads);
int NeverEndingLoop(int nThreads);
void NetFlowPacketParser(char *packet, struct sockaddr *from, int packetsize);

/*********************/
/* Flag Functions    */
/*********************/
int IsTrashLoggingEnabled();
int IsTextLoggingEnabled();
int IsBinaryLoggingEnabled();
int IsDetached();

/*********************/
/* Storage Names     */
/*********************/
char *GetBinaryLogFileName();
char *GetTextLogFileName();
char *GetTrashLogFileName();
void ReleaseNamedVars();
void CommitData();

/*********************/
/* Storage Functions */
/*********************/
void InitStorage();
void CloseStorage();
int StoreNetFlowPacket(struct FlowItem *i);
void LogWrite(char *message, char *file);
void TextLogPacket(struct FlowItem *i);

#ifdef HAVE_LIBSQLITE
void CreateDB(sqlite *db);
void SQLWrite(struct FlowItem *i);
void RotateDB();
#endif

#ifdef HAVE_LIBMYSQLCLIENT
int IsMYSQLLoggingEnabled();
char *GetMYSQLServerName();
void MySQLWrite(struct FlowItem *i);
#endif

void CreateSnapshot();
int  FileCopy(char *src, char *dst);


/*********************/
/* Utility functions */
/*********************/
void IpProtocol(unsigned char proto, char *strprotocol);
void HelpCommandLine();
void Terminate(int error,char *caller);
void HaltApp();

/*********************/
/* Signal Management */
/*********************/
void SwitchTrashFlag();
void SwitchTextFlag();
void HandleSnapshotSignal();
void SetupSignals();
void EnableWaitState();
void DisableWaitState();
void StartShutdown();
int ShutdownInProgress();

#ifdef ENABLE_PTHREADS
void SetWaitTime(int nThreads);
void SetThreadCount(int nThreads);
void SetJobQueue();
void SetThrottle(int condition);
int  IsReceiving();
unsigned long GetMaxThrottle();

int  AliveThreads();
void InitThreadList();
void AddThread(pthread_t tid, int thread_type, int thread_status, int index);
void EndThread(pthread_t tid,int index);
pthread_t FindThread(int thread_type);

void InitQueueCounters();
unsigned long GetProcessedCount();
unsigned long GetQueueCount();
unsigned long GetReceivedCount();
void IncrementProcessed();
void IncrementQueued();
void IncrementReceived();
void ResetQueued();

#endif


unsigned long GetStorageWrittenCount();
unsigned long GetStorageFreedCount();


#ifdef DEBUG
void ShowDebugger();
#endif

#define ERR_OK   	0
#define ERR_NOMEM	1
#define ERR_FILEREAD 	2
#define ERR_FILEWRITE	3
#define ERR_FILEACCESS	4
#define ERR_THREADCOUNT	5
#define ERR_CANTBIND	6
#define ERR_THREADCREAT	7
#define ERR_CANTUSEDB   8
#define ERR_SHUTDOWN	9
#define ERR_CANTOPEN_RO	10
#define ERR_CANTOPEN_RW	11
#define ERR_NOSPACE	12
#define ERR_MYSQLCONN	13
#define ERR_MYSQLWRITE	14


#endif
