/* $Id: parser.c,v 1.1.1.1 2004/11/06 01:16:48 x86man Exp $ */

#include "config.h"
#include "neyed.h"

void NetFlowPacketParser(char *packet, struct sockaddr *from,int packetsize)
{
	unsigned short version;
	unsigned short srcas;
	unsigned short dstas;
	char *tmpstring;
	char address[IPSTRSIZE];
	char srcip[IPSTRSIZE];
	char dstip[IPSTRSIZE];
	char nexthop[IPSTRSIZE];
	char strstart[TIMESTRSIZE];
	char strstop[TIMESTRSIZE];
	char epoch[TIMESTRSIZE];
	char strnow[TIMESTRSIZE];
	char strprotocol[STDSTRSIZE];
	char protocol;
	char message[STDSTRSIZE];
#ifdef DEBUG
	char ErrMessage[STDSTRSIZE];
#endif

	time_t dstart, dend, now;
	
	long uptime,unixseconds,unixnanoseconds,sequence;
	unsigned int sport,dport,port;
	
	NetFlowHeader 	header;
	NetFlowEntry  	entry;
	FlowItem	*item;
	
	
	struct sockaddr_in *cvt;
	
	
	int flowcount,i;

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in netflowparser function");
#endif	
#endif
#endif
#endif

	time(&now);

	cvt = (struct sockaddr_in *)from;	
	inet_ntop(AF_INET,&cvt->sin_addr,(char *)&address,IPSTRSIZE);

	port = ntohs(cvt->sin_port);

	tmpstring = (char *)ctime((time_t *)&now);
	if (tmpstring != NULL)
		sprintf(strnow,"%s",tmpstring);
	
	if (packetsize < sizeof(NetFlowHeader) )
	{
		if (IsTrashLoggingEnabled())
		{
			sprintf(message,"Origin: %s - Warning useless packet of %d bytes with sport: %d at %s",address,packetsize,port,strnow);
			LogWrite(message,(char *)GetTrashLogFileName());
		}
		return;
	}

	memcpy(&header,packet,sizeof(NetFlowHeader));

	version= ntohs(header.version);	
	switch(version)
	{
		case 5:
		{
			flowcount =ntohs(header.count);
			unixseconds = ntohl(header.unixseconds);
			unixnanoseconds = ntohl(header.unixnanoseconds);
			sequence = ntohl(header.sequence);
			

			if (flowcount != ( (packetsize - sizeof(header)) / sizeof(entry) ))
			{
				if (IsTrashLoggingEnabled())
				{
					sprintf(message,"Origin: %s - Incoming NETFLOW V%d Mangled packet at %s",address,version,strnow);
					LogWrite(message,(char *)GetTrashLogFileName());
				}
				return;
			}

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
			sprintf(ErrMessage,"Origin: %s - NETFLOW V%d packet on port %d (contains %d flows) at %s",
						address, 
						version,
						port,
						flowcount,
						strnow);
			trace("=========== PDU DECODE =====================");
			trace(ErrMessage);
#endif
#endif
#endif

			for (i=0;i<flowcount;i++)
			{
				memcpy(&entry,packet+sizeof(header)+(i*sizeof(entry)),sizeof(entry));
				inet_ntop(AF_INET,&entry.s_ip,(char *)&srcip,IPSTRSIZE);
					
				inet_ntop(AF_INET,&entry.d_ip,(char *)&dstip,IPSTRSIZE);
			
				inet_ntop(AF_INET,&entry.next_hop,(char *)&nexthop,IPSTRSIZE);
					
				sport = ntohs(entry.src_port);
				dport = ntohs(entry.dst_port);
				uptime = ntohl(header.uptime);
				dstart = unixseconds + (ntohl(entry.time_start)/1000);
				dend   = unixseconds + (ntohl(entry.time_stop)/1000);				
				protocol = entry.ip_protocol;
				
/*				IpProtocol(protocol,strprotocol); */
				
				
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG				
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - From: %s/%d To: %s/%d NEXT HOP:%s\n",
								address,
								i+1,
								srcip,
								sport,
								dstip,
								dport,
								nexthop);
				trace(ErrMessage);
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - Protocol: %d bytes: %ld packets:%ld Sequence: %ld\n",
								address,
								i+1,
								protocol,
								(unsigned long)ntohl(entry.bytes),
								(unsigned long)ntohl(entry.packets),
								sequence);
				trace(ErrMessage);

				sprintf(ErrMessage,"uptime: %ld uxseconds: %ld uxns: %ld dstart: %ld, dend: %ld\n",uptime,unixseconds,unixnanoseconds,dstart,dend);
				trace(ErrMessage);
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - TCP Flags:0x%x - TOS: 0x%x\n",
						address,
						i+1,
						entry.tcp_flags,
						entry.tos);
				trace(ErrMessage);						
#endif
#endif
#endif										
				srcas = ntohs(entry.src_as);
				dstas = ntohs(entry.dst_as);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - Src AS: %d Dst AS: %d\n",address,i+1, srcas, dstas);
				trace(ErrMessage);
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - Src Mask: %d Dst Mask: %d\n",address,i+1,entry.src_netmask,entry.dst_netmask);
				trace(ErrMessage);				
#endif
#endif
#endif

				tmpstring = (char *)ctime((time_t *)&dstart);
				if (tmpstring != NULL)				
					sprintf(strstart,"%s",tmpstring);

				tmpstring = (char *)ctime((time_t *)&dend);
				if (tmpstring != NULL)				
					sprintf(strstop,"%s",tmpstring);
								
				tmpstring = (char *)ctime((time_t *)&unixseconds);
				if (tmpstring != NULL)				
					sprintf(epoch,"%s",tmpstring);
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - Uptime    : %ld\n",address,i+1,uptime);
				trace(ErrMessage);
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - Started on: %s",address,i+1,strstart);
				trace(ErrMessage);
				sprintf(ErrMessage,"Origin: %s - Flow N.%d - Finish  on: %s",address,i+1,strstop);
				trace(ErrMessage);				
#endif
#endif
#endif

				item = (struct FlowItem *)allocate_mem(sizeof(FlowItem));

				item->origin_ip = 	(unsigned long)cvt->sin_addr.s_addr;
				item->time_received = 	now;
				item->version =		version;
				item->sequence =	sequence;
				item->src_ip =		(unsigned long)entry.s_ip;
				item->src_port =	sport;
				item->dst_ip =		(unsigned long)entry.d_ip;
				item->dst_port =	dport;
				item->next_hop =	(unsigned long)entry.next_hop;
				item->protocol =	protocol;
				item->tcp_flags =	entry.tcp_flags;
				item->tos =		entry.tos;
				item->bytes =		entry.bytes;
				item->packets =		entry.packets;
				item->flow_begin =	dstart;
				item->flow_end =	dend;
				item->src_as =		srcas;
				item->dst_as =		dstas;
				item->src_mask =	entry.src_netmask;
				item->dst_mask =	entry.dst_netmask;

				StoreNetFlowPacket(item);
				release_mem(item,sizeof(struct FlowItem));
				/* item = NULL; */
			}
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
			
			trace("=========== DECODE COMPLETE =================");
#endif
#endif
#endif

			
			break;
		
		}
		default:
		{
			if (IsTrashLoggingEnabled())
			{
				sprintf(message,"Origin: %s - Version %d unsupported packet at %s",address,version,strnow);
				LogWrite(message,(char *)GetTrashLogFileName());
			}
			break;
		}
	}

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of netflowparser function");
#endif	
#endif
#endif
#endif
	return;
}
