/* $Id: sched.c,v 1.1.1.1 2004/11/06 01:16:48 x86man Exp $ */

#include "config.h"
#include "neyed.h"

#ifdef HAVE_LIBPTHREAD

#include "jobs.h"

static int TotalThreads = 0;
static int WaitTime = 0;
static int WaitStates = 1;

static long timelostinws = 0L;
static long predictionhits = 0L;
static long predictionfails = 0L;

struct JobItem *head = NULL;

/*************************************
This is the latest cacheing stuff
introduced
*************************************/
struct JobItem *LastItem = NULL;
int IsLastItemValid = 0;

#ifndef OSX
pthread_mutex_t PacketList_mutex = PTHREAD_MUTEX_INITIALIZER;; 
/* pthread_mutex_t PacketList_mutex = PTHREAD_MUTEX_ERRORCHECK; */
#else
pthread_mutex_t PacketList_mutex = PTHREAD_MUTEX_INITIALIZER;; 
#endif


static int under_pressure;

int IsReceiving()
{
	return under_pressure;
}

void SetThrottle(int condition)
{
	if ((under_pressure == 1) && condition == 1)
	{
		/* We're throttling too much */
		if ( (GetReceivedCount() - GetProcessedCount()) > GetMaxThrottle()) 
			under_pressure = 0;
	}
	else under_pressure = condition;
}


static int next_job_thread = 0;

int GetNextThread()
{
	if (next_job_thread < (TotalThreads-1))
	{
		next_job_thread++;
	}
	else
	{
		next_job_thread = 0;
	}					
	return next_job_thread;
}


typedef struct thread_matrix {
	pthread_t	thread_id;
	int		thread_type;
	int		thread_status;
	int 		index;
} thread_matrix;

static struct thread_matrix TID[MAX_THREADS];

void AddThread(pthread_t tid, int thread_type, int thread_status,int index)
{
	int i;

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in AddThread function");
#endif
#endif
#endif
#endif
	for (i=0;i < MAX_THREADS; i++)
	{
		if (TID[i].thread_id == 0)
		{
			TID[i].thread_id     = tid;
			TID[i].thread_type   = thread_type;
			TID[i].thread_status = thread_status;
			TID[i].index	     = index;
			return;
		}
	}


#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of AddThread function");
#endif
#endif
#endif
#endif

}

void InitThreadList()
{
	int i;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in InitThreadList function");
#endif
#endif
#endif
#endif
	
	for (i=0;i < MAX_THREADS; i++)
	{
		TID[i].thread_id = 0;
		TID[i].thread_type = THREAD_TYPE_UNDEFINED;
		TID[i].thread_status = THREAD_UNDEFINED;
		TID[i].index = -1;
	}
	
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of InitThreadList function");
#endif
#endif
#endif
#endif
	
}

pthread_t FindThread(int thread_type)
{
	int i;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in FindThread function");
#endif
#endif
#endif
#endif
	for (i=0;i < MAX_THREADS; i++)
	{
		if (TID[i].thread_type == THREAD_TYPE_UNDEFINED)
			break;
		if (TID[i].thread_type == thread_type)
			return TID[i].thread_id;
	}
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of FindThread function");
#endif
#endif
#endif
#endif

	return 0;
}

void EndThread(pthread_t tid,int index)
{
	int i;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in EndThread function");
#endif
#endif
#endif
#endif
	for (i=0;i < MAX_THREADS; i++)
	{
		if (TID[i].thread_type == THREAD_TYPE_UNDEFINED)
			break;
			
		if (tid == 0)
		{
			if (TID[i].index == index)
			{
				TID[i].thread_status = THREAD_ENDED;
				break;
			}
		}
		else
		{
			if (TID[i].thread_id == tid)
			{
				TID[i].thread_status = THREAD_ENDED;
				break;
			}
		}
	}	
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of EndThread function");
#endif
#endif
#endif
#endif

}

int AliveThreads()
{
	int count;
	int i;
	
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in AliveThreads function");
#endif
#endif
#endif
#endif
	
	count = 0;
	
	for (i=0;i < MAX_THREADS; i++)
	{
		if (TID[i].thread_type == THREAD_TYPE_UNDEFINED)
			break;
		if (TID[i].thread_status == THREAD_ALIVE)
		{
			count++;
		}
	}	

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of AliveThreads function");
#endif
#endif
#endif
#endif

	return count;	
}


void EnableWaitState()
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in EnableWaitState function");
#endif
#endif
#endif
#endif

	WaitStates = 1;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of EnableWaitState function");
#endif
#endif
#endif
#endif
}
void DisableWaitState()
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in DisableWaitState function");
#endif
#endif
#endif
#endif
	WaitStates = 0;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of DisableWaitState function");
#endif
#endif
#endif
#endif
}

int MinimizeDelay()
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in MinimizeDelay function");
#endif
#endif
#endif
#endif
	return WaitStates;
}

void WaitState()
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in WaitState function");
#endif
#endif
#endif
#endif
	usleep(WaitTime);
	timelostinws+=WaitTime;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of WaitState function");
#endif
#endif
#endif
#endif
}

void YieldCPU()
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in YieldCPU function");
#endif
#endif
#endif
#endif
	sleep(TotalThreads+1);
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of YieldCPU function");
#endif
#endif
#endif
#endif
}

/******************************
This function returns null if
there are no allocated slots
(ie all jobs done)
or the pointer of the LAST 
available slot, where you will
link.....
NB: The list mutex must be already
locked
*****************************/

struct JobItem *FindFirstSlot()
{
	struct JobItem *cycle;
	struct JobItem *prev;
	struct JobItem *RetVal;

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in FindFirstSlot function");
#endif
#endif
#endif
#endif

	if (head == NULL)
	{
		RetVal = NULL;
		return RetVal;
	}

	cycle = head;
	prev = NULL;

	do
	{
		if (cycle == NULL)
		{
			RetVal = prev;
			break;
		}
		if (cycle->next == NULL)
		{
			RetVal = cycle;
			break;
		}
		else 
		{
			prev = cycle;
			cycle = cycle->next;
		}
	} while (cycle != NULL);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of FindFirstSlot function");
#endif
#endif
#endif
#endif

	return RetVal;
}

struct JobItem *FindFirstJob(int threadid)
{
	struct JobItem *cycle;
	struct JobItem *RetVal;
#ifdef DEBUG
	char 	ErrMessage[STDSTRSIZE];
#ifdef VERBOSEDEBUG	
	long    attempts;
#endif	
	
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in FindFirstJob function");
#endif
#endif
#endif
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
	attempts = 0;
#endif
#endif


	while (pthread_mutex_trylock(&PacketList_mutex) !=0 )
	{

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG

		fprintf(stderr,"FindFirstJob: Thread %d waiting to get lock to search for jobs (attempt #%ld)\n",threadid,attempts);
		attempts++;
#endif
#endif
#endif

		WaitState(); 
	}

	RetVal = NULL;

	cycle = head;

	do
	{
		if (cycle == NULL)
		{
			RetVal = NULL;
			break;
		}
		if (cycle->thread_assigned == threadid && cycle->JobComplete == 0)
		{
			RetVal = cycle;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
			if (IsDetached() == 0)
			{
				sprintf(ErrMessage,"Found job for thread %d\n",threadid);
				trace(ErrMessage);
			}
#endif
#endif
#endif
			break;
		}
		else
		{
			cycle = cycle->next;
		}
	} while (1);

	pthread_mutex_unlock(&PacketList_mutex);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of  FindFirstJob function");
#endif
#endif
#endif
#endif

	return RetVal;
}


/*******************************************************************
This works as a sort of Cache to avoid searching in the linked list
every time, specially in stress conditions. It gives a big boost
improvement while a flood of packets is coming into the collector
since avoids walking all the linked list to find the last item
of that one, thus avoiding the loss of time required in the Addjob
function invoked in the listener (udp, no connection == no fork )
*******************************************************************/
struct JobItem *GetLastItem()
{
	struct JobItem *retval;
	
	if (IsLastItemValid == 1)
		retval = LastItem;
	else
	{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
		fprintf(stderr,"Item Cache is invalid\n");
#endif
#endif		
		retval = FindFirstSlot();
	}
	return retval;
}

void AddJob(char *packet,long packetsize,struct sockaddr *from)
{
	struct 	JobItem *n;
	struct 	JobItem *i;
	char 	*packetbuffer;
#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];
	int	locktrial;
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in AddJob function");
#endif
#endif
#endif
#endif

	assert(packet != NULL);
	
	n = (struct JobItem *)allocate_mem(sizeof(struct JobItem));

	n->thread_assigned = GetNextThread();
	n->JobComplete = 0;

	packetbuffer = allocate_mem(packetsize);

	memcpy(packetbuffer,packet,packetsize);
	n->packet = packetbuffer;
	n->packetsize = packetsize;
	n->from = allocate_mem(sizeof(struct sockaddr));
	
	memcpy(n->from,from,sizeof(struct sockaddr));
	n->prev = NULL;
	n->next = NULL;

	/* Lock went here just to speed up things */

#ifdef DEBUG
	locktrial = 0;
#endif	

	while (pthread_mutex_trylock(&PacketList_mutex) !=0 )
	{
		WaitState(); 
#ifdef DEBUG
		sprintf(ErrMessage,"AddJob: Trying to get lock #%d\n",locktrial);
		trace(ErrMessage);
		locktrial++;
#endif				
	}

	i = GetLastItem();

	if (i == NULL)
	{
		/* First item in the list */
		head = n;
		head->prev = NULL;
		head->next = NULL;
		LastItem = head;
		IsLastItemValid = 1;
	}
	else
	{
		n->prev = i;
		i->next = n;
		LastItem = n;
		IsLastItemValid = 1;
	}

	IncrementReceived();

	pthread_mutex_unlock(&PacketList_mutex);


#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG

	if (IsDetached() == 0)
	{
		sprintf(ErrMessage,"Packet of %d bytes added to queue\n",n->packetsize);
		trace(ErrMessage);
	}
#endif		
#endif
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of AddJob function");
#endif
#endif
#endif
#endif

	
}

void ThreadWorker(int threadid)
{
	struct 	JobItem *job;
	long	processedjobs;
#ifdef DEBUG
	char 	ErrMessage[STDSTRSIZE];
#endif
	int acceleratefactor;
	int iteration = 0;
	int step = 0;

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in ThreadWorker function");
#endif
#endif
#endif
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
	if (IsDetached() == 0)
	{
		sprintf(ErrMessage,"Thread %d created\n",threadid);
		trace(ErrMessage);
	}
#endif
#endif
#endif

	acceleratefactor = TotalThreads * 4;
	processedjobs = 0;
	while (1)
	{
	
		if (predictionfails > predictionhits)
		{
			acceleratefactor = max(TotalThreads,acceleratefactor-1);
		}
		else if (predictionhits > predictionfails)
		{
			acceleratefactor++;
		}
	
		job = NULL;
		while (IsReceiving() == 1)
		{
			WaitState();
		}

		/* This is a bottleneck since the whole list is locked and
                   the other threads wait for me to get a single job ptr, 
                   the listener too waits for me. So my idea is, since
		   jobs are assigned round robin, to try to walk on fire
		   and find subsequent (at most TotalThreads) jobs without
		   recurring to locking FindFirstJob */
		job = FindFirstJob(threadid);
		if (job == NULL)
		{
			if (ShutdownInProgress() == 1)
			{	
				EndThread(0,threadid);
				break;
			}
			else
			{
				WaitState();
			}
		}
		for (iteration=0; iteration < acceleratefactor; iteration++)
		{
			if (iteration !=0 )
			{
				for (step=0; step < TotalThreads; step++)
				{
					if (job == NULL)
					{
						predictionfails++;
						break;
					}
					job = job->next;
					if (job == NULL)
					{
						predictionfails++;
						break;
					}
				}
				if (job != NULL)
				{
					if (job->thread_assigned != threadid)
					{
						job = NULL;
						predictionfails++;
						break;
					}
					else predictionhits++;
					
				}
			}
			if (job != NULL)
			{
			/* Here we process the packet */
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
				if (IsDetached() == 0)
				{
					sprintf(ErrMessage,"Thread %ld evaluating job at 0x%p\n",(long)threadid,job);
					trace(ErrMessage);	
				}
#endif
#endif
#endif
			
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
				if (IsDetached() == 0)
				{
					sprintf(ErrMessage,"Thread %ld processing job at 0x%p\n",(long)threadid,job);
					trace(ErrMessage);	
				}
#endif
#endif
#endif
				NetFlowPacketParser(job->packet,
							job->from,
							job->packetsize);
			
			/* No lock required here since the packet will never
                           be freed if JobComplete !=1 and the other threads
			   will never access this job memory area , at least
                           it would work this way ! */	
				processedjobs++;
				job->JobComplete = 1;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
				if (IsDetached() == 0)
				{
					sprintf(ErrMessage,"Thread %d completed processing packet\n",threadid);
					trace(ErrMessage);	
				}
#endif		
#endif
#endif		
#ifdef DEBUG
#ifdef VERBOSEDEBUG
			
		
				if (IsDetached() == 0)
				{
					sprintf(ErrMessage,"Thread %d processed %ld jobs up to now\n",threadid,processedjobs);
					trace(ErrMessage);
				}
#endif
#endif

			}
			else break;
		}
	}

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of ThreadWorker function");
#endif
#endif
#endif
#endif

}

void SetThreadCount(int nThreads)
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in SetThreadCount function");
#endif
#endif
#endif
#endif

	TotalThreads = nThreads;
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of SetThreadCount function");
#endif
#endif
#endif
#endif

}

void SetWaitTime(int nThreads)
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in of SetWaitTime function");
#endif
#endif
#endif
#endif
	WaitTime = 1;

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of SetWaitTime function");
#endif
#endif
#endif
#endif
}

void SetJobQueue()
{

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm in SetJobQueue function");
#endif
#endif
#endif
#endif

	head = NULL;

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of SetJobQueue function");
#endif
#endif
#endif
#endif

}

void JobListMaintainer(int nThreads)
{
	struct 	JobItem *cycle;
	struct 	JobItem *next;
	struct 	JobItem *prev;
		
#ifdef DEBUG
	char 	ErrMessage[STDSTRSIZE];
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{		
		sprintf(ErrMessage,"I'm starting Thread Job Dispatcher with %d threads\n",nThreads);
		trace(ErrMessage);
	}
#endif
#endif
#endif
#endif
	InitQueueCounters();
	pthread_mutex_init(&PacketList_mutex,NULL);

	while (1)
	{

		while (IsReceiving() == 1)
		{
			WaitState();
		}
		
		while (pthread_mutex_trylock(&PacketList_mutex) !=0 )
		{
			WaitState(); 
		}

		cycle = head;

		ResetQueued();

		IsLastItemValid = 0;
		
		do
		{
			
			if (cycle == NULL)
			{
				break;
			}
			else
			{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
				if (IsDetached() == 0)
				{
					sprintf(ErrMessage,"Dispatcher(FREE): Job %ld Thread %ld Completed: %ld\n",(long)GetQueueCount(),(long)cycle->thread_assigned,(long)cycle->JobComplete);
					trace(ErrMessage);
				}
#endif
#endif
#endif
				IncrementQueued();
							
				if (cycle->JobComplete == 1)
				{
					IncrementProcessed();
					
					/* will need to store the job, free it and
						rearrange queue */
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
					if (IsDetached() == 0)
					{
						sprintf(ErrMessage,"=======>  Freeing job for thread %ld  <============\n",(long)cycle->thread_assigned);
						trace(ErrMessage);
					}
#endif
#endif
#endif
					/* Here we free the buffer and reset
                                           stuff in the linked list ... */
					   
					next = cycle->next;
					prev = cycle->prev;
					
					if (cycle == head)
					{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
						if (IsDetached() == 0)
						{
							trace("Dispatcher: HEAD FREED !");
						}
#endif
#endif
#endif
						release_mem(cycle->from,sizeof(struct sockaddr));
						release_mem(cycle->packet,cycle->packetsize);
						release_mem(cycle,sizeof(struct JobItem));
						
						head = cycle = next;
						if (next == NULL)
							break;
					}
					else
					{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
						if (IsDetached() == 0)
						{
							trace("Dispatcher: FREEING ELEMENT !");
						}
#endif						
#endif
#endif					
						if (prev != NULL)
						{
							prev->next = next;
						}
					
						if (next != NULL)
						{
							next->prev = prev;
							
						}

						release_mem(cycle->from,sizeof(struct sockaddr));
						release_mem(cycle->packet,cycle->packetsize);
						release_mem(cycle,sizeof(struct JobItem));

						
						if (next != NULL)
							cycle = next;
						else
							break;
						
					}
						
				}
				else
				{
					cycle = cycle -> next;
				}
			}
		} while (cycle != NULL);


		pthread_mutex_unlock(&PacketList_mutex);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
		if (IsDetached() == 0)
		{
			sprintf(ErrMessage,"Dispatcher: Searching (%ld PDU recvd ,%ld freed, %ld encountered)...\n",
						GetReceivedCount(),
						GetProcessedCount(),
						GetQueueCount());
			trace(ErrMessage);
		}
#endif
#endif		
		YieldCPU(); 

		if (ShutdownInProgress() == 1 && GetQueueCount() == 0)
		{
			EndThread(FindThread(THREAD_TYPE_DISPATCHER),0);
			break;
		}
	}
#ifdef DEBUG
		fprintf(stderr,"WaitState/Lock Time (uS) : %ld - Predictions hit: %ld  - Predictions failed %ld\r",timelostinws,predictionhits,predictionfails);
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	trace("I'm just out of the Thread Dispatcher function");
#endif
#endif
#endif
#endif

#ifdef DMALLOC
	dmalloc_log_stats();
#endif

}

#endif
