/* $Id: storage.c,v 1.1.1.1 2004/11/06 01:16:48 x86man Exp $ */

#include "config.h"
#include "neyed.h"

#ifdef HAVE_LIBSQLITE
static sqlite *sqlitedb;
static int commit_queue = 0;
#endif

#ifdef HAVE_LIBMYSQLCLIENT
MYSQL *mysqlconn;
MYSQL mysql;
#endif

static unsigned long PacketsToStore = 0;
static unsigned long PacketsFreed = 0;

unsigned long GetStorageWrittenCount()
{
	return PacketsToStore;
}

unsigned long GetStorageFreedCount()
{
	return PacketsFreed;
}

#ifdef HAVE_LIBPTHREAD

typedef struct JobStorage
{
	struct FlowItem   *data;
	struct JobStorage *prev;
	struct JobStorage *next;
} JobStorage;

struct JobStorage *LastJob = NULL;
int IsLastJobValid = 0;


pthread_mutex_t StoreData_mutex = PTHREAD_MUTEX_INITIALIZER;

struct JobStorage *datastore = NULL;

#endif

#ifdef HAVE_LIBSQLITE

int FileCopy(char *source, char *dest)
{
	int RetVal;
	FILE *fdsrc;
	FILE *fddst;
	void *buffer;
	size_t effectiveread;
	size_t effectivewrite;

#ifdef DEBUG
	char ErrMessage[STDSTRSIZE];
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in FileCopy function");
	}
#endif
#endif
#endif
#endif
		
	RetVal = ERR_OK;
	
	chmod(source,S_IRUSR);
		
	fdsrc = fopen(source,"rb");
	if (fdsrc == NULL)
	{
		return ERR_CANTOPEN_RO;
	}

	fddst = fopen(dest,"w+b");
	if (fddst == NULL)
	{
		fclose(fdsrc);
		return ERR_CANTOPEN_RW;
	}

	chmod(source,S_IWUSR);
	
	buffer = allocate_mem(COPYBUFSIZE);
	
	do
	{
		effectiveread = fread(buffer,1,COPYBUFSIZE,fdsrc);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
		
		if (IsDetached() == 0)
		{
			sprintf(ErrMessage,"Filecopy: Read %d bytes",effectiveread);	
			trace(ErrMessage);
		}
#endif
#endif
#endif
		if (effectiveread > 0)
		{
			effectivewrite = fwrite(buffer,1,effectiveread,fddst);
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
			if (IsDetached() == 0)
			{
				sprintf(ErrMessage,"Filecopy: Wrote %d bytes",effectivewrite);		
				trace(ErrMessage);
			}
#endif
#endif
#endif
			if ( effectivewrite != effectiveread)
			{
				RetVal = ERR_NOSPACE;
				break;
			}
		}
		else 
		{
			if (feof(fdsrc) !=0 )
			{
				RetVal = ERR_OK;
			}
			else
			{
				RetVal = ERR_FILEREAD;
			}
			break;
		}
		
	} while(1);
	
	fclose(fddst);
	fclose(fdsrc);
	release_mem(buffer,COPYBUFSIZE);
	sync();

	chmod(source,S_IRUSR|S_IWUSR);
	chmod(dest,S_IRUSR|S_IWUSR);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("Filecopy: complete");
	}
#endif
#endif
#endif
#endif
	return RetVal;
}

#endif

#ifdef HAVE_LIBSQLITE

void CreateDB(sqlite *db)
{
	char *errmsg = NULL;
	int RetVal;

#ifdef DEBUG
	char ErrMessage[STDSTRSIZE];
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in CreateDB function");
	}
#endif
#endif
#endif
#endif

	RetVal = sqlite_exec(db,"CREATE TABLE flows (RECNUM INTEGER PRIMARY KEY,PROBE CHAR,TIME_RECEIVED INTEGER, FLOW_VERSION INTEGER, SEQUENCE INTEGER, SOURCE_IP CHAR, SOURCE_PORT INTEGER, DEST_IP CHAR, DEST_PORT INTEGER, NEXT_HOP CHAR, PROTOCOL INTEGER, TCP_FLAGS INTEGER, TOS INTEGER, BYTES INTEGER, PACKETS INTEGER, FLOW_BEGIN INTEGER, FLOW_END INTEGER, SOURCE_AS INTEGER, DEST_AS INTEGER, SOURCE_MASK INTEGER, DEST_MASK INTEGER)", NULL, 0, &errmsg);
	switch (RetVal)
	{
		case SQLITE_OK:
		{
			break;
		}
		case SQLITE_BUSY:
		{
			Terminate(ERR_CANTUSEDB,"The database is locked !\n");
			break;
		}
		case SQLITE_LOCKED:
		{
			Terminate(ERR_CANTUSEDB,"The table is locked !\n");
			break;
		}
		case SQLITE_PERM:
		{
			Terminate(ERR_CANTUSEDB,"No permission to access table !\n");
			break;
		}
		case SQLITE_READONLY:
		{
			Terminate(ERR_CANTUSEDB,"The table is READ ONLY !\n");
			break;
		}
		case SQLITE_INTERNAL:
		{
			Terminate(ERR_CANTUSEDB,"SQLITE internal error !\n");
			break;
		}
		default:
		{
			printf("Err code is: %d\n",RetVal);
			Terminate(ERR_CANTUSEDB,errmsg);
			break;
		}
	}
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of CreateDB function");
	}
#endif
#endif
#endif
#endif

}
#endif


#ifdef HAVE_LIBPTHREAD
int IsStoring()
{
	return IsLastJobValid;
}

/*******************************************************

HERE THE LIST _MUST_ ALREADY BE LOCKED !!!!!!!!!!!!

********************************************************/
struct JobStorage *GetLastJob()
{
	struct 	JobStorage *cycle;
	struct 	JobStorage *prev;

	struct JobStorage *retval;

	retval = NULL;

	if (IsLastJobValid == 1)
	{
		retval=LastJob;
	}
	else
	{

#ifdef DEBUG
#ifdef VERBOSEDEBUG
		fprintf(stderr,"StorageJob Cache is not valid\n");
#endif
#endif		
		
		if (datastore == NULL)
		{
			retval = NULL;
		}
		else
		{
			cycle = datastore;
			prev  = NULL;

			do
			{
				if (cycle == NULL)
				{
					retval = prev;
					break;
				}
				if (cycle->next == NULL)
				{
					retval = cycle;
					break;
				}
				else 
				{
					prev = cycle;
					cycle = cycle->next;
				}				
			} while (cycle != NULL);
		}
	}
	return retval;
}



void AddStorageJob(struct FlowItem *item)
{
	struct 	JobStorage *cycle;
	struct 	JobStorage *prev;
	struct 	JobStorage *next;
	struct 	JobStorage *newitem;
	struct 	JobStorage *last;

	int 	count;

#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];	
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in AddStorageJob function");
	}
#endif
#endif
#endif
#endif
	
	while (pthread_mutex_trylock(&StoreData_mutex) !=0 )
	{
		WaitState(); 
	}

	cycle = prev = next= newitem = (struct JobStorage *) NULL;

	
	count = 0;
	
	newitem = (struct JobStorage *)allocate_mem(sizeof(struct JobStorage));
	if (newitem == NULL)
	{
		Terminate(ERR_NOMEM,"AddStorageJob - create JobStruct");
	}

/*
	memset((void *)newitem,0,sizeof(struct JobStorage));
*/

	newitem->data = (struct FlowItem *)allocate_mem(sizeof(struct FlowItem));
	if (newitem->data == NULL)
	{
		Terminate(ERR_NOMEM,"AddStorageJob - create FlowItem");
	}
	
	memcpy(newitem->data,item,sizeof(FlowItem));
	newitem->next = NULL;
	newitem->prev = NULL;


	last = GetLastJob();	

	if (last == NULL)
	{
		datastore = newitem;
		datastore->next = NULL;
		datastore->prev = NULL;
		LastJob = datastore;
		IsLastJobValid = 1;
	}
	else
	{
		last->next = newitem;
		newitem->prev = last;
		LastJob = newitem;
		IsLastJobValid = 1;
	}

	count++;

	pthread_mutex_unlock(&StoreData_mutex);
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
	if (IsDetached() == 0)
	{
		sprintf(ErrMessage,"ADDJOB: Added %d jobs",count);
		trace(ErrMessage);
	}
#endif
#endif
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of AddStorageJob function");
	}
#endif
#endif
#endif
#endif
}

void StorageJob()
{
	struct JobStorage *cycle;
	struct JobStorage *next;
	int activity;
	
#ifdef DEBUG
	char ErrMessage[STDSTRSIZE];
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in StorageJob function");
	}
#endif
#endif
#endif
#endif

#ifdef HAVE_LIBPTHREAD
	pthread_mutex_init(&StoreData_mutex,NULL);
#endif
	
	InitStorage();
	
	datastore = NULL;
	cycle = next = NULL;
	PacketsFreed = 0;
		
	while(1)
	{
		activity = 0;

		while (pthread_mutex_trylock(&StoreData_mutex) !=0 )
		{
			WaitState(); 
		}
		cycle = datastore;
		next  = NULL;
		
		IsLastJobValid = 0;
		LastJob = NULL;

		while (cycle != NULL)
		{
			activity = 1;
			if (cycle != NULL)
			{
				if (cycle->data != NULL)
				{
					/* Store del pacchetto */
				
					if (IsTextLoggingEnabled() )
					{
						TextLogPacket(cycle->data);
					}

					if (IsBinaryLoggingEnabled() )
					{
#ifdef HAVE_LIBSQLITE
						SQLWrite(cycle->data);
#endif
					}
#ifdef HAVE_LIBMYSQLCLIENT
					if (IsMYSQLLoggingEnabled())
					{
						MySQLWrite(cycle->data);
					}
#endif				
				}
				else
				{
                                        if (IsDetached() == 0)
                                        {
						trace("Warning, data pointer but no data !!");
						next = cycle->next;
						goto airbag;
					}
				}
				
				next = cycle->next;
				
/* There is a strange segfault here, but it should not happen ! */
				release_mem(cycle->data,sizeof(struct FlowItem));
airbag:
				release_mem(cycle,sizeof(struct JobStorage));
				cycle->data = NULL;
				cycle = NULL; 

				PacketsFreed++;
				
				datastore = next;
				cycle = next;
			}
			else break;
		}

		datastore = NULL;
				
		pthread_mutex_unlock(&StoreData_mutex);
		if (activity == 1)
		{
			sync();	
		}
		sleep(5);
#ifdef DEBUG
#ifdef VERBOSEDEBUG
		if (IsDetached() == 0)
		{
			sprintf(ErrMessage,"%ld packets stored up to now (%ld freed)",PacketsToStore,PacketsFreed);
			trace(ErrMessage);
		}
#endif
#endif
		if (ShutdownInProgress() == 1)
		{
			EndThread(FindThread(THREAD_TYPE_STORAGE),0);
			break;
		}
	}
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of StorageJob function");
	}
#endif
#endif
#endif
#endif

}


#endif

void TextLogPacket(struct FlowItem *i)
{
	char message[STDSTRSIZE];
	char *tmpstring;

	char origin_ip[IPSTRSIZE];
	char src_ip[IPSTRSIZE];
	char dst_ip[IPSTRSIZE];
	char next_hop[IPSTRSIZE];
	char time_received[TIMESTRSIZE];
	char flow_start[TIMESTRSIZE];
	char flow_stop[TIMESTRSIZE];
	char strprotocol[STDSTRSIZE];

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in TextLogPacket function");
	}
#endif
#endif
#endif
#endif

	assert(i!= NULL);

	/* Here we dump the text file */
	tmpstring = (char *)inet_ntop(AF_INET,&i->origin_ip,(char *)&origin_ip,IPSTRSIZE);
	assert(tmpstring != NULL);
	tmpstring = (char *)inet_ntop(AF_INET,&i->src_ip,(char *)&src_ip,IPSTRSIZE);
	assert(tmpstring != NULL);
	tmpstring = (char *)inet_ntop(AF_INET,&i->dst_ip,(char *)&dst_ip,IPSTRSIZE);
	assert(tmpstring != NULL);
	tmpstring = (char *)inet_ntop(AF_INET,&i->next_hop,(char *)&next_hop,IPSTRSIZE);
	assert(tmpstring != NULL);

	memset(time_received,0,sizeof(time_received));
	memset(flow_start,0,sizeof(flow_start));		
	memset(flow_stop,0,sizeof(flow_stop));

	tmpstring = (char *)ctime(&i->time_received);
	if (tmpstring != NULL)
	{
		sprintf(time_received,"%s",tmpstring);
		time_received[strlen(time_received)-1] = 0;
	}
	else sprintf(time_received,"Unknown");

	tmpstring = (char *)ctime(&i->flow_begin);
	if (tmpstring != NULL)
	{
		sprintf(flow_start,"%s",tmpstring);
		flow_start[strlen(flow_start)-1] = 0;
	}
	else sprintf(flow_start,"Unknown");

	tmpstring = (char *)ctime(&i->flow_end);
	if (tmpstring != NULL)
	{
		sprintf(flow_stop,"%s",tmpstring);
		flow_stop[strlen(flow_stop)-1] = 0;
	}
	else sprintf(flow_stop,"Unknown");

/*	IpProtocol(i->protocol,strprotocol); */

	sprintf(message,"Origin: %15s - Received: %s - Version: %d Sequence: %10ld From: %15s Sport: %5d To: %15s Dport: %5d Next_Hop: %15s Protocol: %5d TCP_Flags: 0x%02x TOS:0x%02x Bytes: %12ld Packets: %12ld Start: %s Stop: %s SrcMask: /%02d DstMask: /%02d\n",
			origin_ip,
			time_received,
			i->version,
			i->sequence,
			src_ip,
			i->src_port,
			dst_ip,
			i->dst_port,
			next_hop,
			i->protocol,
			i->tcp_flags,
			i->tos,
			(unsigned long)ntohl((unsigned long)i->bytes),
			(unsigned long)ntohl((unsigned long)i->packets),
			flow_start,
			flow_stop,
			i->src_mask,
			i->dst_mask);
	LogWrite(message,(char *)GetTextLogFileName());

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of TextLogPacket function");
	}
#endif
#endif
#endif
#endif

}

void LogWrite(char *message, char *file)
{
        FILE *log;
	
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in LogWrite function");
	}
#endif
#endif
#endif
#endif	
	


	umask(066);

	log = fopen(file,"a+");
	if (log == NULL)
	{
		Terminate(ERR_FILEWRITE,"LogWrite - file open");
	}

        fprintf(log,"%s",message);
	fclose(log);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of LogWrite function");
	}
#endif
#endif
#endif
#endif	

	
}



int StoreNetFlowPacket(struct FlowItem *i)
{

#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];	
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in StoreNetFlowPacket function");
	}
#endif
#endif
#endif
#endif	

	
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
	if (IsDetached() == 0)
	{
		sprintf(ErrMessage,"Will have to store %d bytes",sizeof(struct FlowItem));
		trace(ErrMessage);
	}
#endif
#endif
#endif

#ifdef HAVE_LIBPTHREAD
	AddStorageJob(i);
#else


	if (IsTextLoggingEnabled() )
	{
		TextLogPacket(i);
	}
	
	if (IsBinaryLoggingEnabled() )
	{
#ifdef HAVE_LIBSQLITE
		SQLWrite(i);
#endif
	}
#ifdef HAVE_LIBMYSQLCLIENT
	if (IsMYSQLLoggingEnabled() )
	{
		MySQLWrite(i);
	}
#endif	
#endif

	PacketsToStore++;

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of StoreNetFlowPacket function");
	}
#endif
#endif
#endif
#endif	

	return 0;
}

#ifdef HAVE_LIBSQLITE

void SQLWrite(struct FlowItem *i)
{
	char SQLCODE[2048];
	char *tmpstring;
	char *errmsg = 0;
	int RetVal;

	char origin_ip[IPSTRSIZE];
	char src_ip[IPSTRSIZE];
	char dst_ip[IPSTRSIZE];
	char next_hop[IPSTRSIZE];

#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];	
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in SQLWrite function");
	}
#endif
#endif
#endif
#endif	


	/* Here we dump the text file */
	tmpstring = (char *)inet_ntop(AF_INET,&i->origin_ip,(char *)&origin_ip,IPSTRSIZE);
	tmpstring = (char *)inet_ntop(AF_INET,&i->src_ip,(char *)&src_ip,IPSTRSIZE);
	tmpstring = (char *)inet_ntop(AF_INET,&i->dst_ip,(char *)&dst_ip,IPSTRSIZE);
	tmpstring = (char *)inet_ntop(AF_INET,&i->next_hop,(char *)&next_hop,IPSTRSIZE);


	if (commit_queue == 0)
	{
		sqlite_exec(sqlitedb,"BEGIN TRANSACTION",NULL,0,NULL);
	}

	sprintf(SQLCODE,"INSERT INTO flows VALUES (NULL,'%s',%ld,%d,%ld,'%s',%d,'%s',%d,'%s',%d,%d,%d,%ld,%ld,%ld,%ld,%d,%d,%d,%d)",
			origin_ip,
			i->time_received,
			i->version,
			i->sequence,
			src_ip,
			i->src_port,
			dst_ip,
			i->dst_port,
			next_hop,
			i->protocol,
			i->tcp_flags,
			i->tos,
			(unsigned long)ntohl((unsigned long)i->bytes),
			(unsigned long)ntohl((unsigned long)i->packets),
			(long int)i->flow_begin,
			(long int)i->flow_end,
			i->src_as,
			i->dst_as,
			i->src_mask,
			i->dst_mask);
	do
	{
		RetVal = sqlite_exec(sqlitedb,SQLCODE,NULL,0,&errmsg);
		if (RetVal == SQLITE_LOCKED || RetVal == SQLITE_BUSY)
		{
#ifdef HAVE_LIBPTHREAD
			WaitState();
#endif
		}
		if (RetVal != SQLITE_OK)
		{
			fprintf(stderr,"SQL was: %s",SQLCODE);
			Terminate(ERR_CANTUSEDB,errmsg);
		}
		if (RetVal == SQLITE_OK)
			break;
	} while(1);

	if (commit_queue >= COMMIT_TRANSACTIONS)
	{
		sqlite_exec(sqlitedb,"END TRANSACTION",NULL,0,NULL);
		commit_queue = 0;
		sqlite_exec(sqlitedb,"COMMIT",NULL,0,NULL);
	}
	else commit_queue++;
	
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of SQLWrite function");
	}
#endif
#endif
#endif
#endif	
	
}

#endif

void InitStorage()
{
#if defined(HAVE_LIBMYSQLCLIENT) || defined(HAVE_LIBSQLITE)
	char *errmsg = NULL;
#endif
#ifdef HAVE_LIBMYSQLCLIENT
	char *servername;
#endif
#ifdef HAVE_LIBSQLITE
	char *filename;
	int RetVal;
#endif
#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];	
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in InitStorage function");
	}
#endif
#endif
#endif
#endif	


#ifdef HAVE_LIBSQLITE
	if (IsBinaryLoggingEnabled() == 1)
	{
		filename = GetBinaryLogFileName();
		errmsg= NULL;
		sqlitedb = sqlite_open(filename, 0 , &errmsg);
		if (sqlitedb == 0)
		{
			Terminate(ERR_CANTUSEDB,"Can't open/create database");
		}

		if (errmsg != NULL)
		{
			release_mem(errmsg,strlen(errmsg));
			errmsg = NULL;
		}

		RetVal = sqlite_exec(sqlitedb,"SELECT COUNT() FROM flows",NULL,0,&errmsg);
		if (RetVal != SQLITE_OK)
		{
			CreateDB(sqlitedb);
		}

		if (errmsg != NULL)
		{
#ifdef 	DEBUG
		trace(errmsg);
#endif
/*
			release_mem(errmsg,strlen(errmsg));
			errmsg = NULL;
*/
		}
	}

#endif
#ifdef HAVE_LIBMYSQLCLIENT
	mysqlconn = NULL;
	if (IsMYSQLLoggingEnabled() == 1)
	{
		servername = GetMYSQLServerName();
		mysqlconn = mysql_real_connect(&mysql,
						servername,
						NETFLOWUSER,
						0,
						NETFLOWDB, 
						MYSQLPORT,
						0, 
						0);
		if (mysqlconn == NULL)
		{
			errmsg = (char *)allocate_mem(STDSTRSIZE);
			sprintf(errmsg,"%s\n",mysql_error(&mysql));
			Terminate(ERR_MYSQLCONN,errmsg);
		}
		release_mem(servername,strlen(servername));
	}
#endif	
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of InitStorage function");
	}
#endif
#endif
#endif
#endif	

}

void CloseStorage()
{
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in CloseStorage function");
	}
#endif
#endif
#endif
#endif	


#ifdef HAVE_LIBSQLITE
	if (sqlitedb != 0)
	{
		sqlite_close(sqlitedb);
	}
#endif
#ifdef HAVE_LIBMYSQLCLIENT
	if (IsMYSQLLoggingEnabled() == 1)
	{
		if (mysqlconn != NULL)
			mysql_close(mysqlconn);
	}
#endif

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of CloseStorage function");
	}
#endif
#endif
#endif
#endif	

}

void Terminate(int error, char *caller)
{
	char ErrMsg[STDSTRSIZE];
	CloseStorage();
	sprintf(ErrMsg,"Process terminated with value: %d, quitting from %s\n",error,caller);
	fprintf(stderr,ErrMsg);
	fflush(stderr);
	syslog(LOG_DAEMON,ErrMsg);
	ReleaseNamedVars();
#ifdef DMALLOC
	dmalloc_shutdown();
#endif
	exit(error);
}

void HaltApp()
{
	char	ErrMessage[STDSTRSIZE];	

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in HaltApp function");
	}
#endif
#endif
#endif
#endif	

	StartShutdown();
#ifdef HAVE_LIBPTHREAD
	do 
	{
		if (IsDetached() == 0)
		{
			sprintf(ErrMessage,"Alive threads: %d",AliveThreads());
			trace(ErrMessage);
			sleep(1);
		}
	} while (AliveThreads() > 0);
#else
	waitpid(-1,NULL,0);
	sleep(180);
#endif
	Terminate(ERR_OK,"clean shutdown");
}

#ifdef HAVE_LIBSQLITE
void CreateSnapshot()
{
	int  retval;
	char *errmsg = NULL;
	char snapshot[STDSTRSIZE];

#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];	
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in CreateSnapshot function");
	}
#endif
#endif
#endif
#endif	


	notify("Begun creation of DB snapshot");
	
	sprintf(snapshot,"%s.snapshot",GetBinaryLogFileName());
	

#ifdef HAVE_LIBPTHREAD	
	while (pthread_mutex_trylock(&StoreData_mutex) !=0 )
	{
		WaitState(); 
	}
#endif
	sqlite_close(sqlitedb);

	retval = FileCopy(GetBinaryLogFileName(),snapshot);	
	if (retval != ERR_OK)
	{
		Terminate(retval,"Can't create snapshot");
	}

	sqlitedb = sqlite_open(GetBinaryLogFileName(), 0 , &errmsg);
	if (sqlitedb == 0)
	{
		Terminate(ERR_CANTUSEDB,"Can't open/create database");
	}

#ifdef HAVE_LIBPTHREAD	
	pthread_mutex_unlock(&StoreData_mutex);
#endif
	
	notify("Succesfully completed DB File snapshot creation");

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of CreateSnapshot function");
	}
#endif
#endif
#endif
#endif	

	
}
#endif

#ifdef HAVE_LIBSQLITE
void RotateDB()
{
	int  retval;
	char *errmsg = NULL;
	char snapshot[STDSTRSIZE];

#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];	
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in RotateDB function");
	}
#endif
#endif
#endif
#endif	
	notify("Begun rotation of DB file");

	sprintf(snapshot,"%s.previous",GetBinaryLogFileName());
	

#ifdef HAVE_LIBPTHREAD	
	while (pthread_mutex_trylock(&StoreData_mutex) !=0 )
	{
		WaitState(); 
	}
#endif

	sqlite_close(sqlitedb);

	retval = FileCopy(GetBinaryLogFileName(),snapshot);	
	if (retval != ERR_OK)
	{
		Terminate(retval,"Can't create snapshot");
	}

	sqlitedb = sqlite_open(GetBinaryLogFileName(), 0 , &errmsg);
	if (sqlitedb == 0)
	{
		Terminate(ERR_CANTUSEDB,"Can't open/create database");
	}

	retval = sqlite_exec(sqlitedb,"DROP TABLE FLOWS", NULL, 0, &errmsg);
	if (retval == SQLITE_OK)
	{
		CreateDB(sqlitedb);
	}

/* ???? Redoing same stff ???	
	retval = sqlite_exec(sqlitedb,"VACUUM FLOWS", NULL, 0, &errmsg);
	if (retval == SQLITE_OK)
	{
		CreateDB(sqlitedb);
	}
*/

	sync();

#ifdef HAVE_LIBPTHREAD	
	pthread_mutex_unlock(&StoreData_mutex);
#endif

	notify("Completed rotation of DB File");
#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of RotateDB function");
	}
#endif
#endif
#endif
#endif
	
}

#endif


#ifdef HAVE_LIBMYSQLCLIENT

void MySQLWrite(struct FlowItem *i)
{
	char SQLCODE[2048];
	char *tmpstring;
	int RetVal;

	char origin_ip[IPSTRSIZE];
	char src_ip[IPSTRSIZE];
	char dst_ip[IPSTRSIZE];
	char next_hop[IPSTRSIZE];
	char time_received[TIMESTRSIZE];
	char flow_start[TIMESTRSIZE];
	char flow_stop[TIMESTRSIZE];
	
	struct tm *when;

#ifdef DEBUG
	char	ErrMessage[STDSTRSIZE];	
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm in MySQLWrite function");
	}
#endif
#endif
#endif
#endif

	assert(i != NULL);

	/* Here we dump the text file */
	tmpstring = (char *)inet_ntop(AF_INET,&i->origin_ip,(char *)&origin_ip,IPSTRSIZE);
	tmpstring = (char *)inet_ntop(AF_INET,&i->src_ip,(char *)&src_ip,IPSTRSIZE);
	tmpstring = (char *)inet_ntop(AF_INET,&i->dst_ip,(char *)&dst_ip,IPSTRSIZE);
	tmpstring = (char *)inet_ntop(AF_INET,&i->next_hop,(char *)&next_hop,IPSTRSIZE);


#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
	if (IsDetached() == 0)
	{
		trace("MYSQLWrite: BEFORE TIME CONVERSION");
	}
#endif
#endif
#endif

	memset(time_received,0,sizeof(time_received));
	memset(flow_start,0,sizeof(flow_start));		
	memset(flow_stop,0,sizeof(flow_stop));

	when = localtime(&i->time_received);
	strftime(time_received,sizeof(time_received),"%Y-%m-%d %H:%M:%S",when);
	
	when = localtime(&i->flow_begin);
	strftime(flow_start,sizeof(flow_start),"%Y-%m-%d %H:%M:%S",when);

	when = localtime(&i->flow_end);
	strftime(flow_stop,sizeof(flow_stop),"%Y-%m-%d %H:%M:%S",when);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
	if (IsDetached() == 0)
	{
		trace("MYSQLWrite: AFTER TIME CONVERSION");
	}
#endif
#endif
#endif


	sprintf(SQLCODE,"INSERT INTO FLOWS VALUES (NULL,'%s','%s',%d,%ld,'%s',%d,'%s',%d,'%s',%d,%d,%d,%ld,%ld,'%s','%s',%d,%d,%d,%d);",
			origin_ip,
			time_received,
			i->version,
			i->sequence,
			src_ip,
			i->src_port,
			dst_ip,
			i->dst_port,
			next_hop,
			i->protocol,
			i->tcp_flags,
			i->tos,
			(unsigned long)ntohl((unsigned long)i->bytes),
			(unsigned long)ntohl((unsigned long)i->packets),
			flow_start,
			flow_stop,
			i->src_as,
			i->dst_as,
			i->src_mask,
			i->dst_mask);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
	if (IsDetached() == 0)
	{
		sprintf(ErrMessage,"MySQL Error: %s",SQLCODE);
		trace(ErrMessage);;
	}
#endif
#endif
#endif

	do
	{
		RetVal = mysql_query(mysqlconn,SQLCODE);
		if (RetVal != 0)
		{
			Terminate(ERR_MYSQLWRITE,"Can't add data to mysql server");
		}
		if (RetVal == 0)
			break;
	} while(1);

#ifdef DEBUG
#ifdef VERBOSEDEBUG
#ifdef PARANOIDDEBUG
#ifdef FUNCTRACE
	if (IsDetached() == 0)
	{
		trace("I'm just out of MySQLWrite function");
	}
#endif
#endif
#endif
#endif


}

#endif

void CommitData()
{
#ifdef HAVE_LIBSQLITE
	if (IsBinaryLoggingEnabled() )
	{
/*
		sqlite_exec(sqlitedb,"COMMIT",NULL,0,NULL);
*/
	}
#endif

#ifdef HAVE_LIBMYSQLCLIENT
	if (IsMYSQLLoggingEnabled())
	{
		mysql_query(mysqlconn,"COMMIT");
	}
#endif
	sync();
}
