/* trafgen.c,v 1.1.1.1 2004/07/22 22:44:50 x86man Exp */

#include "config.h"
#include "platform.h"
#include "structs.h"

char *OSVersion()
{
        return strdup(OSNAME);
}


void MakeRandomIP(char *target)
{
	int i1,i2,i3,i4;
	i1 = 1+(int) (223.0*rand()/(RAND_MAX+1.0));
	i2 = 1+(int) (254.0*rand()/(RAND_MAX+1.0));
	i3 = 1+(int) (254.0*rand()/(RAND_MAX+1.0));
	i4 = 1+(int) (254.0*rand()/(RAND_MAX+1.0));
	sprintf(target,"%d.%d.%d.%d",i1,i2,i3,i4);
}

void Usage()
{
	fprintf(stdout,"Usage: trafgen -d destination_ip [ -c num_packets -f -r]\n");
	fprintf(stdout,"-r = fixed rate of about 500 packets per second\n");
	fprintf(stdout,"-f = fixed SRC=1.1.1.1 DST=2.2.2.2 RTR=1.1.1.2 SRC/DST PORT=80\n");
	exit(1);
}

int main(int argc, char **argv)
{
	char *largepacket;
	long largesize;
	char *target = NULL;
	int count=0;
	int arg,i;
	int sd;
	int fixedip;
	int rate;
	
	time_t start,from;
	char srcip[IPSTRSIZE];
	char dstip[IPSTRSIZE];
	char rtrip[IPSTRSIZE];

	struct sockaddr_in addr;

	NetFlowHeader h;
	NetFlowEntry  e;

	fixedip = 0;
	rate = 0;

	while((arg=getopt(argc,argv,"d:c:frh")) != EOF )
	{
		switch(arg)
		{
			case 'h':
			{
				Usage();
				break;
			}
			case 'f':
			{
				fixedip = 1;
				break;
			}
			case 'r':
			{
				rate = 1;
				break;
			}
			case 'd':
			{
				target = strdup(argv[optind-1]);
				break;
			}
			case 'c':
			{
				count = atoi(argv[optind-1]);
				break;
			}
			default:	
				break;
		}
	}

	fprintf(stderr,"Netflow random traffic generator\n");	
        fprintf(stdout,"License: BITGPL, by Gilberto Persico\n");
#ifdef TELCO_SUX
        fprintf(stdout,"Yep, running this software I agree that Italian telcos are suxing thieves !\n");
#else
        #error "You are faking license, thief !"
#endif
	fprintf(stdout,"Running on: %s\n",OSVersion());
	if (target == NULL)
	{
		Usage();
	}

	sd = socket(PF_INET, SOCK_DGRAM, 0);

	bzero(&addr,sizeof(addr));

	addr.sin_family = AF_INET;
	addr.sin_port = htons(2055);
	addr.sin_addr.s_addr = inet_addr(target);

	if (connect(sd,&addr,sizeof(addr)) != 0)
	{
		printf("Can't open socket !\n");
		exit(2);
	}
	
	largesize = sizeof(h) + sizeof(e);
	largepacket = malloc(largesize);

	i=0;


	do
	{

		h.version = htons(5);
		h.count = htons(1);
		h.uptime = 0;
		h.unixseconds= time(NULL);
		h.unixnanoseconds= 0;
		h.sequence = htonl(i);
		h.EngineType = 0;
		h.EngineId = 0;
		h.SampleRate = 0;

		if (fixedip == 0)
		{
			MakeRandomIP(srcip);		
			MakeRandomIP(dstip);		
			MakeRandomIP(rtrip);		
		}
		else
		{
			sprintf(srcip,"1.1.1.1");
			sprintf(dstip,"2.2.2.2");
			sprintf(rtrip,"1.1.1.2");
		}

		e.s_ip = inet_addr(srcip);
		e.d_ip = inet_addr(dstip);
		e.next_hop = inet_addr(rtrip);
		e.inIfIndex = 0;
		e.outIfIndex = 0;
		e.packets = htonl(300000);
		e.bytes=htonl(5000000);
		e.time_start = h.unixseconds + htonl(count/1000);
		e.time_stop = h.unixseconds + htonl((count*2)/1000);
		e.src_port = htons(80);
		if (fixedip == 0)
		{
			e.dst_port = htons(32000+(int) (60000.0*rand()/(RAND_MAX+1.0)));
		}
		else
		{
			e.dst_port = htons(80);
		}
		e.pad = 0;
		e.tcp_flags = 0;
		e.ip_protocol = 6;
		e.tos = 0;
		e.src_as = 0;
		e.dst_as = 0;
		e.src_netmask = 32;
		e.dst_netmask = 32;

		memcpy(largepacket,&h,sizeof(h));
		memcpy(largepacket+sizeof(h),&e,sizeof(e));
	
		send(sd,largepacket,largesize,NULL);

		i++;
		fprintf(stderr,"Sending Packet (size: %ld) %ld\r",largesize,i);
		if (i >= count && count != 0)
			break;

		if (rate ==1 )
			usleep(500); 
	} while (1);

	free(largepacket);
	
	fprintf(stderr,"\n");

	close(sd);
}
